@REM Build and execute C++ test application using Optima
cd test\app
cmake -S . -B build -DCMAKE_PREFIX_PATH=%LIBRARY_PREFIX%
cmake --build build --config Release
build\Release\app.exe




if not exist %PREFIX%\\Library\\include\\eigen3 exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %PREFIX%\\Library\\share\\pkgconfig\\eigen3.pc exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
